"""
此脚本用来提取高斯输出文件的信息
高斯的输出文件包含迭代信息(结构优化、扫描等)
但是封装成分子对象之后就只有一个信息了
所以迭代信息只能在reader对象中存在,且不是默认属性
"""

import re
import numpy as np
from pathlib import Path
from functools import lru_cache
from collections import namedtuple
import threading

from pywfn.data.basis import Basis
from pywfn.maths.gto import Gto
from pywfn.reader.lutils import Reader
from pywfn.utils import printer
from pywfn.data.elements import elements

from typing import Callable
import linecache

class Title:
    def __init__(self,mark:str,jtype:int=0,multi:bool=False) -> None:
        self.line:int=-1
        self.mark:str=mark
        self.jtype=0 # 0：正则表达式匹配，1：包含式匹配，2：全等匹配
        self.multi=multi # 是否持续查找（多个中查找最后一个）
    
    def judge(self,line:str):
        """判断所给行是否满足条件"""
        if self.jtype==0:
            return re.search(self.mark,line) is not None
        elif self.jtype==1:
            return self.mark in line

    
    def __repr__(self) -> str:
        return f'{self.line}: {self.mark}'

class LogReader(Reader):
    def __init__(self, path:str):
        super().__init__(path)
        assert path[-4:] in ['.log','.out'],'文件类型不匹配，应为.log文件或.out文件'
        self.index=0 #从第一行向下搜索，搜索到需要的就停止
        self.titles={ #记录每一个title所在的行数
            'coords':Title(r'(Input orientation|Standard orientation)',0,True),
            'basis':Title('Standard basis:',1),
            'coefs':Title(r'  Molecular Orbital Coefficients',0),
            'acoefs':Title(r'Alpha Molecular Orbital Coefficients',1),
            'bcoefs':Title(r'Beta Molecular Orbital Coefficients',1),
            'overlap':Title('\*\*\* Overlap \*\*\*',1),
            'kinetic':Title('\*\*\* Kinetic Energy \*\*\*',1),
            'potential':Title('\*\*\* Potential Energy \*\*\*',1),
            'density':Title(r'Density Matrix:',1),
            'basisData':Title(r'Overlap normalization',1),
            'engs':Title(r'Zero-point correction',1),
            'keyWards':Title(r'# .+',0),
        }
        self.search_title()
    
    @property
    def normalEnd(self)->bool:
        lastLine=self.getline(self.lineNum)
        return 'Normal termination of Gaussian' in lastLine
    
    @lru_cache
    def get_coords(self)->np.ndarray:
        """原子坐标[n,3]"""
        symbols,coords=self.read_coords()
        assert isinstance(coords,np.ndarray),'coord必须是np.ndarray类型'
        return coords

    @lru_cache
    def get_symbols(self)->list[str]:
        """原子符号[n]"""
        symbols,coords=self.read_coords()
        return symbols

    @lru_cache
    def get_CM(self) -> np.ndarray:
        CM=self.read_CMs()[-1]
        return CM
    
    @lru_cache
    def get_obtAtms(self) -> list[int]:
        ObtAtms=self.read_CMs()[0]
        return ObtAtms
    
    @lru_cache
    def get_obtShls(self) -> list[int]:
        ObtShls=self.read_CMs()[1]
        return ObtShls


    @lru_cache
    def get_obtSyms(self)-> list[list[int]]:
        obtSyms=self.read_CMs()[2]
        return obtSyms

    @lru_cache
    def get_obtEngs(self) -> list[float]:
        ObtEngs=self.read_CMs()[3]
        return ObtEngs
    
    def get_obtOccs(self) -> list[bool]:
        ObtOccs=self.read_CMs()[4]
        return ObtOccs

    @lru_cache
    def get_SM(self)->np.ndarray:
        return self.read_SM()
    
    def get_charge(self) -> int:
        charge,spin=self.read_multiy()
        return charge
    
    def get_spin(self)->int:
        charge,spin=self.read_multiy()
        return spin
    
    def get_energy(self)->float:
        return self.read_energy()
    
    from pywfn.data import Basis
    def get_basis(self)->Basis:
        name=self.read_basisName()
        data=self.read_basisData()
        basis=Basis(name)
        basis.setData(data)
        return basis


    def search_title(self):
        """
        搜索到需要用到的标题行就停止
        """
        # 所有标题所在的行
        def sear_group(start:int): #每一个搜索的线程
            for j in range(start,start+100_000):
                line=self.getline(j)
                if line=='':break
                for key in self.titles.keys():
                    title:Title=self.titles[key]
                    if title.line!=-1 and title.multi==False:continue
                    if title.judge(line):
                        self.titles[key].line=j
                        # print(j,line)
        threads:list[threading.Thread]=[]
        for i in range(0,self.lineNum,100_000):
            t=threading.Thread(target=sear_group,args=(i,))
            t.start()
            threads.append(t)
        for t in threads:
            t.join()

    def read_keyWrds(self):
        """读取关键字"""
        titleNum=self.titles['keyWards'].line
        keyWards=''
        for i in range(titleNum,titleNum+3):
            line=self.getline(i,False).strip()
            if line=='-'*70:break
            keyWards+=line
        return keyWards


    @lru_cache
    def read_coords(self)->tuple[list[str],np.ndarray]:
        '''读取原子坐标'''
        
        titleNum=self.titles['coords'].line
        assert titleNum!=-1,"没有坐标对应的行"
        if titleNum is None:
            printer.warn('没有读取到原子坐标')
            return
        s1=r' +\d+ +(\d+) +\d +(-?\d+.\d{6}) +(-?\d+.\d{6}) +(-?\d+.\d{6})'
        coords=[]
        symbols=[]
        for i in range(titleNum+5,self.lineNum):
            line=self.getline(i)
            if re.search(s1, line) is not None:
                res=list(re.search(s1, line).groups())
                atomID=int(res[0])
                symbol=elements[atomID].symbol
                coord=[float(each) for each in res[1:]]
                symbols.append(symbol)
                coords.append(coord)
            else:
                return symbols,np.array(coords)*1.889

    @lru_cache
    def read_multiy(self):
        """读取电荷和自选多重度"""
        res=re.findall(f'Charge = +(-?\d) Multiplicity = (\d)',self.text)
        if res is None:
            printer.warn(f'{self.path} 没有读到电荷和自选多重度!')
        else:
            charge,multiy=res[0]
            return int(charge),int(multiy)

    @lru_cache 
    def read_basisName(self)->str:
        """读取基组名"""
        titleNum=self.titles['basis'].line
        if titleNum is None:
            printer.warn('未读取到基组名')
            name='unll'
        else:
            name=re.match(' Standard basis: (\S+) ',self.getline(titleNum)).group(1)
        return name
    
    @lru_cache
    def read_basisData(self):
        """
        读取基组数据
        原子类型
            电子层
        """
        from pywfn.data.basis import BasisData
        assert 'gfinput' in self.read_keyWrds(),'关键词应该包含：gfinput'
        titleNum=self.titles['basisData'].line
        symbols=self.get_symbols()
        if titleNum is None:return
        basisDatas:list[BasisData]=[]
        ifRead=True
        angDict={'S':0,'P':1,'D':2} #角动量对应的字典
        s1='^ +(\d+) +\d+$'
        s2=r' ([SPD]+) +(\d+) \d.\d{2} +\d.\d{12}'
        s3=r'^ +(( +-?\d.\d{10}D[+-]\d{2}){2,3})'
        s4=' ****\n'
        atomics=[] # 已经获取过的元素
        atomic=None
        shell=0
        angs=None
        exp=None
        coe=None
        for i in range(titleNum+1,self.lineNum):
            line=self.getline(i)
            if re.search(s1,line) is not None:
                idx=re.search(s1,line).groups()[0] #第几个原子
                idx=int(idx)-1
                symbol=symbols[idx]
                atomic=elements[symbol].idx

                if atomic not in atomics:
                    atomics.append(atomic) #shells
                    ifRead=True # 该元素是否已经读过
                else:
                    ifRead=False
            elif re.search(s2,line) is not None:
                if not ifRead:continue
                shellName,lineNum=re.search(s2,line).groups()
                angs=[angDict[s] for s in shellName] #角动量
                shell+=1
            elif re.search(s3,line) is not None:
                if not ifRead:continue
                numsStr=re.search(s3,line).groups()[0]
                nums:list[str]=re.findall(r'-?\d.\d{10}D[+-]\d{2}',numsStr)
                nums=[float(num.replace('D','E')) for num in nums]
                if len(angs)==1:
                    exp,coe=nums
                    data=BasisData(atomic,shell,angs[0],exp,coe)
                    basisDatas.append(data)
                if len(angs)==2:
                    exp,coe1,coe2=nums
                    data1=BasisData(atomic,shell,angs[0],exp,coe1)
                    data2=BasisData(atomic,shell,angs[1],exp,coe2)
                    basisDatas.append(data1)
                    basisDatas.append(data2)
            elif line==s4 is not None:
                shell=0
                continue
            else:
                # 排序一下
                basisDatas.sort(key=lambda b:(b.atmic,b.shl,b.ang))
                return basisDatas
        
    def read_summery(self):
        '''读取总结信息'''
        summerys=re.findall(r' \d[\\||]\d[\S\s]*?@',self.content)
        summery=summerys[-1]
        summery=''.join([each[1:] for each in summery.split('\n')])
        summeryList=summery.replace('\n','').replace('\\\\','||').split('||')
        Infos,KeyWords,JobTitle,Coords=summeryList[:4]
        _,_,_,jobType,method,basisSet,molFormula,user,date,_=Infos.replace('\\','|').split('|')
        summery={
            'basisSet':basisSet,
            'Coords':np.array([[float(num.replace(' ','')) for num in each.split(',')[1:]] for each in Coords.replace('\\','|').split('|')[1:]])
        }
        return summery
           
    # 分子轨道的文本数据有5种情况
    #情况1                           1         2         3         4         5
    #情况2                           O         O         O         O         O
    #情况3     Eigenvalues --   -11.17917 -11.17907 -11.17829 -11.17818 -11.17794
    #情况4   1 1   C  1S         -0.00901  -0.01132   0.00047  -0.01645  -0.02767
    #情况5   2        2S         -0.00131  -0.00175  -0.00041  -0.00184  -0.00173
    @lru_cache
    def read_CM_(self, title:str):  # 提取所有原子的轨道 自己写的代码自己看不懂真实一件可悲的事情,此函数逻辑复杂，要好好整明白
        s1='^(( +\d+){1,5}) *$'
        s2=r'^(( *(\(\w+\)--)?[OV]){1,5})$'
        s3=r'^ +Eigenvalues --(( +-?\d+.\d+){1,5})'
        s4='^ +\d+ +(\d+) +([A-Za-z]+) +(\d[A-Z]+)(( *-?\d+.\d+){1,5})$'
        # s5='^ +\d+ +(\d+[A-Za-z ]+)(( *-?\d+.\d+){1,5})$'
        s5='^ +\d+ +(\d+[A-Za-z]+ ?\+?-?\d?)(( *-?\d+.\d+){1,5})$'

        titleNum=self.titles[title].line
        if titleNum is None:return None

        OrbitalAtom=[] # 系数矩阵每一行对应的原子
        OrbitalLayer=[] # 系数矩阵每一行对应的壳层符号
        OrbitalEngs=[] # 轨道本征值
        OrbitalType=[] # 轨道类型
        dataDict:dict[int:list[list[float]]]={}
        firstShow=True
        if titleNum is None:
            return
        for i in range(titleNum+1,self.lineNum):
            line=self.getline(i)
            if re.search(s1, line) is not None: #情况1
                pass
            elif re.search(s2, line) is not None: # 情况2，获得column
                OrbitalType += re.split(r' +', line.replace('\n',''))[1:] # 获取占据轨道还是非占据轨道
                
            elif re.search(s3, line) is not None: # 情况3，每个轨道本征值     
                line_data,_=re.search(s3,line).groups()
                line_data=re.findall('-?\d+.\d+', line_data)
                line_data=[float(each) for each in line_data]
                OrbitalEngs+=line_data
            elif re.search(s4,line) is not None: # 原子出现
                atomIDX,atomType,layer,line_data,_=re.search(s4,line).groups()
                atomIDX = int(atomIDX)
                nums=[float(each) for each in re.findall('-?\d+.\d+',line_data)]
                if atomIDX not in dataDict.keys():
                    dataDict[atomIDX]=[]
                else:
                    firstShow=False
                dataDict[atomIDX].append([])
                dataDict[atomIDX][-1].append(nums)
                if firstShow:OrbitalLayer.append(layer)
                if firstShow:OrbitalAtom.append(atomIDX)
                # self.OCdict[atomIDX].set(layer,nums)
            elif re.search(s5,line) is not None:
                layer,line_data,_=re.search(s5,line).groups()
                line_data=re.findall('-?\d+.\d+', line_data)
                nums=[float(each) for each in line_data]
                dataDict[atomIDX][-1].append(nums)
                if firstShow:OrbitalLayer.append(layer)
                if firstShow:OrbitalAtom.append(atomIDX)
                # self.OCdict[atomIDX].set(layer,nums)
            else: # 若不满足以上任意一种情况，说明已经查找完毕，则对收集到的数据进行处理
                printer.console.log(f'读取完成,i={i},line={line}')
                for atomic,matrics in dataDict.items():
                    for i,matrix in enumerate(matrics):
                        dataDict[atomic][i]=np.array(matrix)
                for atomic,matrics in dataDict.items():
                    dataDict[atomic]=np.concatenate(matrics,axis=1)
                CM=np.concatenate([m for m in dataDict.values()],axis=0)
                break
        printer.console.log(f'读取完成CM,shape={CM.shape}')
        assert CM.shape[0]==CM.shape[1],"CM需要为正方形矩阵"
        return OrbitalAtom,OrbitalLayer,OrbitalEngs,OrbitalType,CM
    
    @lru_cache
    def read_CM(self, title:str):  # 提取所有原子的轨道 自己写的代码自己看不懂真实一件可悲的事情,此函数逻辑复杂，要好好整明白
        keyWards=self.read_keyWrds()
        assert 'pop=full' in keyWards,'关键词应包含：pop=full'
        find=re.search('NBasis *= *(\d+)',self.text).groups()[0]
        NBasis=int(find)
        NBlock=NBasis//5+(0 if NBasis%5==0 else 1)
        titleNum=self.titles[title].line
        
        if titleNum is None:return None
        blockLen=NBasis+3 #一块数据行数，在log文件的输出中，轨道系数是按照每一块五列来分块的
        # print(titleNum,NBlock,blockLen)
        ObtAtms=[] # 原子轨道对应的原子
        ObtShls=[] # 占据原子的第多少层轨道
        ObtSyms=[] # 原子轨道对应的角动量
        ObtEngs=[] # 分子轨道能量
        ObtOccs=[] # 占据/非占据
        
        CM=np.zeros(shape=(NBasis,NBasis))
        for i,l in enumerate(range(titleNum+1,titleNum+NBlock*blockLen+1,blockLen)):
            occs=self.getline(l+1)[21:71] # 占据类型的起止位置
            occs=[occs[i:i+10].strip() for i in range(0,50,10)]
            occs=[e for e in occs if e!='']
            occs=[True if e[-1]=='O' else False for e in occs]
            ObtOccs+=occs
            engs =re.split(' +',self.getline(l+2)[21:].strip())
            line=self.getline(l+2)
            engs=[float(e) for e in engs]
            ObtEngs+=engs
            
            coefs=[line[21:] for line in self.getlines(l+3,l+3+NBasis)]
            for j,line in enumerate(coefs):
                coef=re.findall('-?\d+.\d+',line)
                coef=[float(e) for e in coef]
                CM[j,i*5:i*5+len(coef)]=coef

            if i==0: #只在第一块记录行信息
                atomID=''
                for l2 in range(l+3,l+blockLen):
                    line=self.getline(l2)[:15] # 行信息(角动量，对应原子)的起止位置
                    match=line[12:].strip()
                    shl,sym=match[0],match[1:]
                    ObtShls.append(int(shl))
                    ObtSyms.append(sym)
                    obtAtom=line[5:9].strip()
                    if obtAtom!='':
                        atomID=int(obtAtom) # 更改当前行的原子
                    ObtAtms.append(atomID)
        return ObtAtms,ObtShls,ObtSyms,ObtEngs,ObtOccs,CM

    @lru_cache
    def read_CMs(self)->tuple[list,list,list,list,list,np.ndarray]:
        """
        获取轨道系数及相关信息
        atms,shls,angs,engs,occs,CM
        """
        if self.titles['coefs'].line!=-1:
            atms,shls,angs,engs,occs,CM=self.read_CM('coefs')
            return atms,shls,angs,engs,occs,CM
        elif self.titles['acoefs'].line!=-1:
            atmsA,shlsA,angsA,engsA,occsA,CMA=self.read_CM('acoefs')
            atmsB,shlsB,angsB,engsB,occsB,CMB=self.read_CM('bcoefs')
            atms=atmsA
            angs=angsA
            shls=shlsA
            engs=engsA+engsB
            occs=occsA+occsB
            CM=np.concatenate([CMA,CMB],axis=1)
            return atms,shls,angs,engs,occs,CM
        else:
            raise ValueError('没有找到轨道系数')

    @lru_cache
    def read_SM(self):
        """读取重叠矩阵"""
        return self.read_Mat('overlap')

    def read_Mat(self,title:str):
        """读取矩阵"""
        s1='^( +\d+){1,5} *$'
        s2=' +\d+( +-?\d.\d{6}D[+-]\d{2}){1,5} *'
        lineDatas:dict[str:str]={}
        titleNum=self.titles[title].line
        if titleNum is None:
            return None
        for i in range(titleNum+1,self.lineNum):
            line=self.getline(i)
            if re.match(s1, line) is not None:
                pass
            elif re.match(s2, line) is not None:
                idx=line[:8]
                if idx not in lineDatas.keys():
                    lineDatas[idx]=''
                nums=line[8:].replace('D','e')
                lineDatas[idx]+=f'{nums} '
            else:
                break
        size=len(lineDatas.keys())
        SM_=np.zeros((size,size))
        for idx,each in lineDatas.items():
            each:str=each
            nums=re.split(' +',each.strip())
            nums=[float(num) for num in nums]
            SM_[int(idx)-1,:len(nums)]=nums
        return np.tril(SM_)+np.tril(SM_,-1).T

    @lru_cache
    def read_charge(self)->float:
        res=re.search(r'Sum of Mulliken charges = +(-?\d.\d{5})',self.text).group(1)
        return float(res)
    
    def read_energy(self)->float:
        engs=re.findall('SCF Done: +E\(.*\) += +(-?\d+.\d+)',self.text)
        if engs:return float(engs[-1])
        return None

    @lru_cache
    def read_energys(self)->list[float]:
        """返回表格中对应的这8个能量，只返回数值"""
        engList = [
            'Zero-point correction', 
            'Thermal correction to Energy', 
            'Thermal correction to Enthalpy',
            'Thermal correction to Gibbs Free Energy', 
            'Sum of electronic and zero-point Energies',
            'Sum of electronic and thermal Energies', 
            'Sum of electronic and thermal Enthalpies',
            'Sum of electronic and thermal Free Energies'
            ]
        searhNum=0
        engDict={e:None for e in engList}
        lineNum=self.titles['engs'].line
        if lineNum is None:
            return engList,None
        # assert lineNum is not None,f"{self.path} 未读取到能量"
        for i in range(lineNum,self.lineNum):
            line=self.getline(i)
            for each in engList:
                res=re.search(f'{each}=\s+(-?\d+\.\d+)',line)
                if res is not None:
                    re.findall('-?\d+\.\d+',line)
                    engDict[each]=res.groups()[0]
                    searhNum+=1
            if searhNum==len(engList):break
        engNums=[float(e) for e in engDict.values()]
        return engNums



"""
记录一些模板文件
"""

gjf=\
"""%chk=<CHK>.chk
# opt freq b3lyp/6-31g(d) gfinput pop=full iop(3/33=1)

Title Card Required

<CHARGE> <MULTI>
<COORD>

"""

si=\
"""<FILENAME>

<ENERGY>

Cartesian coordinate
<COORD>

Variable frequencies
<FREQ>
"""

pes=\
"""<?xml version="1.0" encoding="UTF-8" ?>
<!DOCTYPE CDXML SYSTEM "http://www.cambridgesoft.com/xml/cdxml.dtd" >
<CDXML
 CreationProgram="ChemDraw 20.0.0.41"
 Name="none.cdxml"
 BoundingBox="0 0 0 0"
 WindowPosition="0 0"
 WindowSize="-2147483648 -2147483648"
 WindowIsZoomed="yes"
 FractionalWidths="yes"
 InterpretChemically="yes"
 ShowAtomQuery="yes"
 ShowAtomStereo="no"
 ShowAtomEnhancedStereo="yes"
 ShowAtomNumber="no"
 ShowResidueID="no"
 ShowBondQuery="yes"
 ShowBondRxn="yes"
 ShowBondStereo="no"
 ShowTerminalCarbonLabels="no"
 ShowNonTerminalCarbonLabels="no"
 HideImplicitHydrogens="no"
 LabelFont="3"
 LabelSize="10"
 LabelFace="96"
 CaptionFont="3"
 CaptionSize="10"
 HashSpacing="2.50"
 MarginWidth="1.60"
 LineWidth="0.60"
 BoldWidth="2"
 BondLength="14.40"
 BondSpacing="18"
 ChainAngle="120"
 LabelJustification="Auto"
 CaptionJustification="Left"
 AminoAcidTermini="HOH"
 ShowSequenceTermini="yes"
 ShowSequenceBonds="yes"
 ShowSequenceUnlinkedBranches="no"
 ResidueWrapCount="40"
 ResidueBlockCount="10"
 ResidueZigZag="yes"
 NumberResidueBlocks="no"
 PrintMargins="36 36 36 36"
 MacPrintInfo="0003000001200120000000000B6608A0FF84FF880BE309180367052703FC0002000001200120000000000B6608A0000100000064000000010001010100000001270F000100010000000000000000000000000002001901900000000000600000000000000000000100000000000000000000000000000000"
 ChemPropName=""
 ChemPropFormula="Chemical Formula: "
 ChemPropExactMass="Exact Mass: "
 ChemPropMolWt="Molecular Weight: "
 ChemPropMOverZ="m/z: "
 ChemPropAnalysis="Elemental Analysis: "
 ChemPropBoilingPt="Boiling Point: "
 ChemPropMeltingPt="Melting Point: "
 ChemPropCritTemp="Critical Temp: "
 ChemPropCritPres="Critical Pres: "
 ChemPropCritVol="Critical Vol: "
 ChemPropGibbs="Gibbs Energy: "
 ChemPropLogP="Log P: "
 ChemPropMR="MR: "
 ChemPropHenry="Henry&apos;s Law: "
 ChemPropEForm="Heat of Form: "
 ChemProptPSA="tPSA: "
 ChemPropCLogP="CLogP: "
 ChemPropCMR="CMR: "
 ChemPropLogS="LogS: "
 ChemPropPKa="pKa: "
 ChemPropID=""
 ChemPropFragmentLabel=""
 color="0"
 bgcolor="1"
 RxnAutonumberStart="1"
 RxnAutonumberConditions="no"
 RxnAutonumberStyle="Roman"
 RxnAutonumberFormat="(#)"
>
<colortable>
<color r="1" g="1" b="1"/>
<color r="0" g="0" b="0"/>
</colortable>
<fonttable>
<font id="1" charset="iso-8859-1" name="Arial"/>
</fonttable>
<page
 id="2"
 BoundingBox="0 0 540 719.75"
 HeaderPosition="36"
 FooterPosition="36"
 PrintTrimMarks="yes"
 HeightPages="1"
 WidthPages="1"
>
<fragment
 id="3"
 BoundingBox="[BOUND]"
 Z="6"
>
[NBS]
</fragment>
[TXS]
</page>
</CDXML>"""

mol="""[TITLE]

Created by pywfn
[NATOM][NBOND]  0  0  0  0  0  0  0  0  0    0
[ATOMBLOCK]
[BONDBLOCK]
M  END
"""


from functools import lru_cache,cached_property

from itertools import product
from pywfn.data.basis import Basis
from pywfn.data.elements import Elements


start=\
"""
欢迎使用 pywfn : 基于python的波函数分析工具
文档 https://www.xiaofei911.top/mkdocs/pywfn/
按q 即可退出程序，如无特殊提示，按 Enter 可返回上一级
按照提示输入编号即可轻松使用本程序
"""


"""
轨道分解法+HMO计算π键级
"""
from pywfn.base import Mol,Atom,Bond
import numpy as np
from pywfn import utils
from pywfn import maths
from pywfn.bondprop import Caler

class Calculator(Caler):
    def __init__(self,mol:Mol) -> None:
        self.mol=mol
        self.bond:list[int]=None

    def get_order(self, center:Atom,around:Atom, orbital:int, direction):  # 计算一个轨道的键级
        '''
        计算两个原子间某个轨道的键级
        中心原子，相邻原子，轨道，方向'''
        if around.symbol=='H':
            return 0
        if np.linalg.norm(direction)==0:
            return 0
        As=self.mol.As[orbital]

        centerPos=center.coord
        aroundPos=around.coord
        bondVector=aroundPos-centerPos
        crossVector=np.cross(utils.normalize(bondVector),utils.normalize(direction))
        

        centerTs=center.OC[:,orbital].copy()
        aroundTs=around.OC[:,orbital].copy()
        if np.linalg.norm(centerTs)==0 or np.linalg.norm(aroundTs)==0:
            return 0
        
        centerPsProj=center.get_pProj(direction, orbital) # 先计算投影
        aroundPsProj=around.get_pProj(direction, orbital)

        nx=bondVector
        ny=crossVector
        nz=direction
        centerRes=utils.coordTrans(nx,ny,nz, centerPsProj)
        aroundRes=utils.coordTrans(nx,ny,nz, aroundPsProj)

        centerPZs=[each[-1] for each in centerRes]
        aroundPZs=[each[-1] for each in aroundRes]

        pOrder=sum([cpz*apz/As for cpz,apz in zip(centerPZs,aroundPZs)])*self.mol.oE
        return pOrder
        # return order

    def get_orders(self,center:Atom,around:Atom,orbitals:list[int],direction):
        """计算每两个原子之间的键级"""
        orders=[self.get_order(center,around,orbital,direction) for orbital in orbitals] # 所有的占据轨道都计算键级
        return orders

    def calculate(self) -> dict:
        """指定一个键，计算该键的键级"""
        idx1,idx2=self.bond
        centerAtom=self.mol.atom(idx1)
        aroundAtom=self.mol.atom(idx2)
        self.centerPIdx=[i for i,l in enumerate(centerAtom.layers) if 'P' in l] #原子的序数
        self.aroundPIdx=[i for i,l in enumerate(aroundAtom.layers) if 'P' in l]
        O_obts=self.mol.O_obts
        normal=centerAtom.get_Normal(aroundAtom) # 原子的法向量
        if len(centerAtom.neighbors)==3: # 如果原子有法向量(sp2)
            orders=self.get_orders(centerAtom,aroundAtom,O_obts,normal)
            return sum(orders)
        else: # 如果没有法向量的话，则需要找轨道方向作为基础方向
            # 从高到低计算轨道方向，遇到有垂直于键轴的则作为轨道方向
            orbitalDirection=None
            for o in O_obts[::-1]:
                orbitalDirection=centerAtom.get_obtWay(o)
                if np.linalg.norm(orbitalDirection)==0:
                    continue
                bondDirection=self.mol.bond(centerAtom.idx, aroundAtom.idx).vector
                if maths.vector_angle(orbitalDirection, bondDirection,trans=True)>0.4: # 夹角要很大
                    break
            if orbitalDirection is not None:
                orders1=self.get_orders(centerAtom,aroundAtom,O_obts,orbitalDirection)
                crossDirection=np.cross(orbitalDirection, bondDirection)
                orders2=self.get_orders(centerAtom,aroundAtom,O_obts,crossDirection)
                return sum(orders1),sum(orders2)
    
    def resStr(self)->str:
        order=self.calculate()
        return f'{order}'



"""
定义一些常见的自定义类型，否则在注释的时候都无法区分
"""
import numpy as np

class Coord(np.ndarray):
    def __init__(self,xyz:list[float]) -> None:
        super().__init__(xyz)
        self.dtype=float

class Lmn(np.ndarray):
    def __init__(self,lmn:list[float]) -> None:
        super().__init__(lmn)
        self.dtype=int


"""
挑选pi轨道
"""
import numpy as np


def judge(Ca:np.ndarray,Cb:np.ndarray,As:float):
    """判断一个轨道是否为pi轨道"""
    # 第一步，计算轨道的贡献
    Sa=np.sum(Ca[0]**2)/As
    Sb=np.sum(Cb[0]**2)/As
    if Sa > 0.05 or Sb > 0.05:
        return 0


# 用来管理程序的一些设置
from pathlib import Path
import numpy as np
import json

configPath=Path.cwd()/'config.json'
if not configPath.exists():
    configPath.touch()
    configPath.write_text('{}')

configDict:dict=json.loads(configPath.read_text())

def get_config(key:str,value):
    if key in configDict.keys():
        return configDict[key]
    else:
        return value

def set_config(key:str,value):
    global configDict
    configDict[key]=value
    configPath.write_text(json.dumps(configDict))

BOND_LIMIT=get_config('BOND_LIMIT',1.7) # 判断量原子之间是否成键的长度限制

IF_DEBUG=True # 是否开启debug,控制项目中所有的打印,避免与shell中的print冲突
IF_SHELL=False # 是否在shell中执行

BASE_VECTOR=np.array([0,0,1]) # 标准向量，求原子法向量和轨道方向的时候要与该向量夹角小于90°

IF_ORBITAL_ORDER=False #是否计算键级中每个轨道的成分，尤其是mayer键级拆分成每个轨道的成分很麻烦

RENDER_ATOM_RANGE=20000  # 原子轨道的渲染范围

RENDER_CLOUD_STEP=0.2  # 渲染间隔

RENDER_CLOUD_BORDER=2

BOHR_RADIUS=1.889

TEMPLATE_PATH_GJF=get_config('TEMPLATE_PATH_GJF',None)
TEMPLATE_PATH_SI=get_config('TEMPLATE_PATH_SI',None)

SHOW_PRINT=False

IF_CM_P=False # 是否使用投影轨道


"""
为Fortran函数写python接口，方便调用
"""

import ctypes as ct
from ctypes import c_int, c_double
from pathlib import Path
import numpy as np
import os

os.add_dll_directory(r"D:\program\mingw64\bin")

path = str(Path(__file__).parent / "flib.dll")
flib = ct.CDLL(path)

flib.add_.argtypes = [ct.c_double, ct.c_double, ct.POINTER(ct.c_double)]


def add_(x: float, y: float):
    res = ct.c_double()
    flib.add_(ct.c_double(x), ct.c_double(y), ct.byref(res))
    return res.value


# 抄的例子
flib.sum2_.argtypes = [ct.POINTER(ct.c_double)]
flib.sum2_.restype = ct.c_double


def sum2(a: float):
    a = ct.byref(ct.c_double(a))
    b = flib.sum2_(a)
    return b


# 抄的例子，矩阵运算
flib.double_array_.argtypes = [ct.POINTER(ct.c_double), ct.c_int]


def double_array():
    # Create a double array, pass it to Fotran as a pointer
    x = np.ones((3, 3), order="F")
    x_ptr = x.ctypes.data_as(ct.POINTER(ct.c_double))

    # Call function
    rint = flib.double_array_(x_ptr, ct.c_int(3))
    return x


flib.grid_pos_.argtypes = [ct.c_int, ct.c_int, ct.c_int, ct.POINTER(ct.c_double)]


def grid_pos(Nx: int, Ny: int, Nz: int):
    pos = np.zeros((Nx * Ny * Nz, 3), order="F")
    pos_ptr = pos.ctypes.data_as(ct.POINTER(ct.c_double))
    flib.grid_pos_(ct.c_int(Nx), ct.c_int(Ny), ct.c_int(Nz), pos_ptr)
    return pos


flib.same_array_.argtypes = [ct.c_int, ct.c_int, ct.POINTER(ct.c_double)]

def same_array():
    row, col = 2, 3
    pos = np.array([[1, 2, 3], [4, 5, 6]], dtype=float, order="F")

    pos_ptr = pos.ctypes.data_as(ct.POINTER(ct.c_double))
    flib.same_array_(ct.c_int(row), ct.c_int(col), pos_ptr)
    return pos


flib.gtf_.argtypes = [
    c_double, # alp
    c_int, # np
    ct.POINTER(ct.c_double), # pos
    ct.POINTER(ct.c_double), # r2
    ct.POINTER(ct.c_int), # lmn
    ct.POINTER(ct.c_double), # val
]

def gtf(exp: float, pos: np.ndarray, R2:np.ndarray, lmn: np.ndarray) -> np.ndarray:
    pos_ptr = pos.ctypes.data_as(ct.POINTER(ct.c_double))
    npos = pos.shape[0]
    R2_ptr = R2.ctypes.data_as(ct.POINTER(ct.c_double))
    lmn_ptr = lmn.ctypes.data_as(ct.POINTER(ct.c_int))
    wfn = np.zeros(npos)
    val_ptr = wfn.ctypes.data_as(ct.POINTER(ct.c_double))
    print(R2.shape)
    flib.gtf_(c_double(exp), c_int(npos), pos_ptr, R2_ptr, lmn_ptr, val_ptr)
    return wfn



"""
该模块用来计算原子属性
Mulliken电荷
"""
from pywfn.utils import printer

import numpy as np

class AtomCaler:
    def __init__(self) -> None:
        self.logTip:str=''

    def resStr(self)->np.ndarray:pass

    def logRes(self):
        printer.info(self.logTip)
        printer.res(self.resStr())


from pywfn.shell import Shell

def onShell(shell:Shell):
    printer.info('1. 原子电荷')
    printer.info('2. 原子自旋')
    printer.info('3. 原子能量')
    printer.info('4. 原子活性')
    printer.info('5. 原子活性')
    opt=printer.info('请输入对应选项：')
    if opt=='':return
    if opt=='1':
        from pywfn.atomprop import charge
        caler=charge.Calculator(shell.input.Moles()[0])
        caler.onShell()
    if opt=='2':
        from pywfn.atomprop import spin
        caler=spin.Calculator(shell.input.Moles()[0])
        caler.onShell()
    if opt=='3':
        from pywfn.atomprop import energy
        caler=energy.Calculator(shell.input.Moles()[0])
        caler.onShell()
    if opt=='4':
        from pywfn.atomprop import activity
        caler=activity.Calculator(shell.input.Moles()[0])
        caler.onShell()



"""
该脚本用来生成SI(支持信息)文件,包含
能量
坐标
频率
"""
from pathlib import Path
import re

from pywfn.base import Mol
from pywfn.utils import printer
from pywfn.data import temps
from pywfn.reader import LogReader

class Tool:
    def __init__(self,mol:Mol) -> None:
        """
        读取文件貌似只需要log/out文件就行了
        """
        self.mol=mol
        self.reader:LogReader=mol.reader #已经实例化的
        self.content=self.reader.text
        self.template=temps.si
        self.selects:list[int]=[1,2,3]
        self.sameFile:bool=True
    
    def write_energy(self):
        engList,engNums=self.reader.read_energys()
        """匹配并保存各种校正能量"""
        ENERGY='\n'.join([f'{name:<45}{value:>15}' for name,value in zip(engList,engNums)])
        self.template=self.template.replace('<ENERGY>',ENERGY)
            
    def write_coord(self):
        COORDS=[]
        for atom in self.mol.atoms:
            symbol=atom.symbol
            x,y,z=atom.coord
            COORDS.append(f'{symbol:4} {x:14.8f} {y:14.8f} {z:14.8f}')
        self.template=self.template.replace('<COORD>','\n'.join(COORDS))

    def write_freq(self):
        freqObj = re.findall(r'^\s+Frequencies\s--\s+(-?\d+\.\d+.+)$', self.content, flags=re.M)
        if freqObj is not None:
            matchContent=f''
            for freq in freqObj:
                freqs=re.split(r' +',freq)
                matchContent+=f'{freqs[0]:>15}{freqs[1]:>15}{freqs[2]:>15}\n'
            self.template=self.template.replace('<FREQ>',matchContent[:-1])
        else:
            printer.wrong('Match Frequencies Error')

    def save(self):
        """
        导出分子的各种信息到txt文件
        1. 坐标 coord
        2. 能量 energy
        3. 频率 freq
        """
        if 1 in self.selects:
            self.write_coord()
        else:
            self.template=self.template.replace('<COORD>\n','')
        if 2 in self.selects:
            self.write_energy()
        else:
            self.template=self.template.replace('<ENERGY>\n','')
        if 3 in self.selects:
            self.write_freq()
        else:
            self.template=self.template.replace('<FREQ>\n','')

        path_=Path(self.reader.path)
        if self.sameFile:
            path=f'{path_.parent}/SI.txt'
            mode='a'
            self.template+=('-'*60+'\n')
        else:
            path=f'{path_.parent}/{path_.stem}_SI.txt'
            mode='w'
        self.template=self.template.replace('<FILENAME>',path_.stem)
        with open(path,mode,encoding='utf-8') as f:
            f.write(self.template)





"""
将多个link的log文件拆分出来
"""
from pathlib import Path
from pywfn.utils import printer
class Tool:
    def __init__(self,path:str) -> None:
        self.path=Path(path)
        self.texts=self.path.read_text()
        self.lines=self.texts.splitlines(keepends=False)
        self.mark=' Initial command:'
        self.nums=[]
        for i,line in enumerate(self.lines):
            if line!=self.mark:continue
            self.nums.append(i)
        self.nums.append(len(self.lines))
    
    def split(self):
        for i in range(len(self.nums)-1):
            l1,l2=self.nums[i:i+2]
            text='\n'.join(self.lines[l1:l2])
            self.write(text,i)
        printer.res('导出完成 >_<')

    def write(self,text,idx):
        name=self.path.stem
        (self.path.parent/f'{name}_{idx+1}.log').write_text(text)
        


"""
计算原子的反应活性，福井函数、parr函数等，一般需要多个分子才行
"""
from pywfn.base import Mol
from pywfn.atomprop import charge, direction, spin
from pywfn.atomprop.charge import Chrgs
import numpy as np
from pywfn.shell import Shell

class Calculator:
    def __init__(self) -> None:
        self.mols:list[Mol]=[]

    def fukui(self,chrg:Chrgs='mulliken')->np.ndarray:
        """计算所有原子的福井函数[n,2]"""
        assert len(self.mols)==3,"应该有三个分子"
        cals=[charge.Calculator(mol) for mol in self.mols]
        natm=len(self.mols[0].atoms)
        chgs=np.zeros(shape=(natm,3)) # 记录所有原子的电荷
        dchg=np.zeros(shape=(natm,2)) # 记录电荷差值
        for c,cal in enumerate(cals):
            res=cal.calculate(chrg)
            chgs[:,c]=res
        dchg[:,0]=-(chgs[:,0]-chgs[:,1])
        dchg[:,1]=-(chgs[:,1]-chgs[:,2])
        return dchg
    
    def parr(self,chrg:Chrgs='mulliken')->np.ndarray:
        """计算所有原子的parr函数[n,2]"""
        assert len(self.mols)==2,"应该有两个分子"
        cals=[spin.Calculator(mol) for mol in self.mols]
        natm=len(self.mols[0].atoms)
        spins=np.zeros(shape=(natm,3))
        for c,cal in enumerate(cals):
            spins[:,c]=cal.calculate(chrg)
        result=np.zeros(shape=(natm,2))
        result[:,0]=spins[:,0]-spins[:,1]
        result[:,1]=spins[:,1]-spins[:,2]
        return spins
    
    def delEng(self):
        from pywfn.atomprop import energy
        cals=[energy.Calculator(mol) for mol in self.mols]
        natm=len(cals[0].mol.atoms)
        engs=np.zeros(shape=(natm,3))
        for c,cal in enumerate(cals):
            res=cal.calculate()
            engs[:,c]=res
        result=np.zeros(shape=(natm,2))
        result[:,0]=engs[:,0]-engs[:,1]
        result[:,1]=engs[:,1]-engs[:,2]
        return engs

    def valence(self)->list[float]:
        """计算原子化合价"""
        assert len(self.mols)==1,"只能算一个分子"
        from pywfn.bondprop import bondOrder
        mol=self.mols[0]
        caler=bondOrder.Calculator(mol)
        orders=caler.mayer()
        orderDict={}
        for a1,a2,order in orders:
            a1,a2=int(a1),int(a2)
            key=f'{a1}-{a2}' if a1<a2 else f'{a2}-{a1}'
            orderDict[key]=order
        result=np.zeros(len(mol.atoms))
        for a,atom in enumerate(mol.atoms):
            a1=atom.idx
            valence=0
            for a2 in atom.neighbors:
                key=f'{a1}-{a2}' if a1<a2 else f'{a2}-{a1}'
                valence+=orderDict[key]
            result[a]=valence
        return result

    def freeValence(self,atms:list[int]):
        """计算指定原子的自由价[d,5](atm,x,y,z,val)"""
        assert len(self.mols)==1,"只能算一个分子"
        from pywfn.bondprop import bondOrder
        mol=self.mols[0]
        caler=bondOrder.Calculator(mol)
        # STAND=1.6494
        STAND=4.0
        result=[]
        for atm1 in atms:
            dirs=None
            orders=[]
            for atm2 in mol.atom(atm1).neighbors:
                bond=[atm1,atm2]
                res=caler.dirMayer([bond]) #[d,4](a1,a2,x,y,z,v)
                if dirs is None:dirs=res[:,2:5]
                vals=res[:,5]
                orders.append(vals)
            orderSum=np.array(orders).sum(axis=0)
            for d in range(len(dirs)):
                x,y,z=dirs[d]
                order=STAND-orderSum[d]
                result.append([atm1,x,y,z,order])
        return np.array(result)

    def dirFukui(self,atms:list[int])->np.ndarray:
        """计算指定原子，指定方向的福井函数[n,6](atm,x,y,z,E,N)"""
        assert len(self.mols)==3,"需要三个分子"
        crgs=[mol.charge for mol in self.mols]
        assert crgs[0]<crgs[1]<crgs[2],"电荷顺序不符"
        cals=[charge.Calculator(mol) for mol in self.mols]
        
        vals:list[np.ndarray]=[]
        dirs=None
        idxs=None
        for cal in cals:
            res=cal.dirCharge('mulliken',atms)
            if dirs is None:dirs=res[:,1:4]
            if idxs is None:idxs=res[:,0]
            vals.append(res[:,4])
        num=len(dirs)
        result=np.zeros(shape=(num,6))
        result[:,0]=idxs
        result[:,1:4]=dirs
        result[:,4]=vals[0]-vals[1]
        result[:,5]=vals[1]-vals[2]
        return result
    
    
    def onShell(self,shell:Shell):
        from pywfn.utils import printer
        opts=[
            'fukui函数',
            'parr函数',
            '电荷差值',
            '化合价',
            '自由价',
            '方向fukui函数'
        ]
        for i,each in enumerate(opts):
            printer.info(f'{i+1}.{each}')
        opt=input('请选择要计算的活性:')
        if opt=='1':
            self.mols=shell.input.Moles()
            printer.info('请输入电荷类型：')
            printer.info('1.mulliken; 2.lowdin')
            chrg=input()
            result=self.fukui()


"""
计算亲核亲电能
"""
from pywfn.base import Mol
from pywfn.atomprop import energy

import numpy as np

class Calculator:
    def __init__(self,moln:Mol,mol0:Mol,molp:Mol) -> None:
        self.mols=[moln,mol0,molp]
        self.cals=[energy.Calculator(m) for m in self.mols]
        self.natm=len(mol0.atoms)
        self.engs=np.zeros((self.natm,3))
        self.deng=np.zeros((self.natm,2))
    
    def calculate(self):
        for c,cal in enumerate(self.cals):
            engs=cal.calculate()
            self.engs[:,c]=engs
        self.deng[:,0]=self.engs[:,0]-self.engs[:,1]
        self.deng[:,1]=self.engs[:,1]-self.engs[:,2]
        return self.deng


# 将需要的每个模块共同需要的计算都放在这里面
import re
import numpy as np

from pywfn import config


def parse_atmList(string)->list[int]:
    """将字符串解析为数字列表"""
    res = []
    for each in re.split(r',|，', string):
        content = each.split('-')
        if len(content) == 1:
            res.append(int(each) - 1)
        else:
            res += [int(i) - 1 for i in range(int(content[0]), int(content[1]) + 1)]
    return res

def get_vertical(v1,v2):
    '''获得垂直于两向量的单位向量'''
    v=np.cross(v1,v2)
    return v/np.linalg.norm(v)

def get_points_between_two_pos(pos1,pos2,n):
    '''计算两点之间的点的坐标'''
    dp=pos2-pos1
    all_pos=[pos1+dp/(n-1)*i for i in range(n)]
    return np.array(all_pos)


def coordTrans(nx,ny,nz,coords):
    """对一个向量实行空间坐标变换
    nx,ny,nz: 另一组基坐标
    coord:要尽心变换的坐标/向量
    """
    e=np.array([[1,0,0],[0,1,0],[0,0,1]]).T # 空间坐标变换
    nx,ny,nz=normalize(nx),normalize(ny),normalize(nz)
    e_=np.array([nx,ny,nz]).T
    A_=np.dot(np.linalg.inv(e_),e)
    ps__=np.array([np.dot(A_,p) for p in coords])
    return ps__

arounds=np.array([
        [-1,0,0],
        [+1,0,0],
        [0,-1,0],
        [0,+1,0],
        [0,0,-1],
        [0,0,+1]
    ])

def differ_function(posan1,posan2): #计算电子分布差值图
    return (posan1+posan2)**2-(posan1**2+posan2**2)/2


def get_gridPoints(range,step,ball=False): 
    '''获取空间格点[3,n]'''
    points=[]
    for x in np.arange(-range,range,step):
        for y in np.arange(-range,range,step):
            for z in np.arange(-range,range,step):
                if ball:
                    distance=(x**2+y**2+z**2)**0.5
                    if distance>range:
                        continue
                points.append([x,y,z])
    return np.array(points).T

def k_means(des,points):
    '''定义k-means聚类算法函数,返回每一类对应的索引'''
    N=len(des)
    for epoch in range(5):
        types=[[] for n in range(N)]
        idxs=[[] for n in range(N)]
        for _,each in enumerate(points):
            distance=np.linalg.norm((des-each),axis=1) #计算每个点与中心点的距离(范数)，两个数值
            idx=np.argmin(distance) # 获取距离最近的索引
            types[idx].append(each) # 在索引位置添加点（分类）
            idxs[idx].append(_)
        types=[np.array(each) for each in types]
        pos=np.array([np.mean(each,axis=0) if len(each)>0 else des[i] for i,each in enumerate(types)]) # 获取两个类别的中心点坐标
        des=pos  # 更新目标点
    return idxs

def list_remove(l,x):
    new_l=[]
    for each in l:
        if each!=x:
            new_l.append(each)
    return new_l

def connectH(atom,connections):
    '''
    如果连接的有H,就返回C-H键的向量,否则返回None
    '''
    CHVectors=[]
    for each in connections:
        if each['atom_type']=='H':
            CHVectors.append(each['pos']-atom['pos'])
    return CHVectors

def multiple(a,b):
    '''获取两个量之间的倍数关系，必定是大于等于1的'''
    if abs(a)>=abs(b):
        return a/b
    else:
        return b/a

def nodeNum(data):
    '''获取节点曲线中的节点数量，有可能会找多'''
    data[np.where(data==0)]=1e-6
    res=data[:-1]*data[1:]
    return len(np.where(res<=0)[0])

def normalize(vector):
    length=np.linalg.norm(vector)
    if length==0:
        raise
    return vector/length

from rich.console import Console
from rich .table import Table,box
from typing import Sequence,Iterable
from pywfn import config
import inspect
class Printer:
    def __init__(self) -> None:
        self.ifDebug=config.IF_DEBUG
        self.ifShell=config.IF_SHELL
        self.console=Console()
        self.tables:dict[str,Table]={}

    def __call__(self,text,style='',end='\n'):
        if self.ifShell:self.console.print(text,style=style,end=end)

    def warn(self,text):
        self.__call__(text,style='#fcc419')

    def info(self,text):
        self.__call__(text,style='#91a7ff')

    def wrong(self,text):
        self.__call__(text,style='#e03131')
    
    def res(self,text):
        self.__call__(text,style='#a9e34b')
        self.bar()
    
    def print(self,text):
        self.__call__(text)
    
    def multi(self,texts:list[str]):
        self.__call__(text,style='')
        for text in texts:
            self.console.print(text,end='')
        self.console.print('')
    
    def bar(self,len=40):
        self.__call__('-'*len,style='#6741d9')

    def space(self):
        self.__call__('')
    
    def log(self,text):
        if self.ifDebug: # 只有debug模式启动的时候才会打印
            frame=inspect.stack()[1]
            text=f'{frame.function}|{frame.filename}:{frame.lineno}\n{text}'
            self.__call__(text,style='#228be6')
    
    def vector(self,tip:str,vector:np.ndarray):
        if not config.IF_SHELL:return
        vector=[f'{v:.2f}' for v in vector]
        vectorStr=','.join(vector)
        self.console.print(tip,end='',style='#dee2e6')
        self.console.print(vectorStr,style='#f59f00')
    
    def options(self,title,opts:dict[str,str]):
        if not config.IF_SHELL:return
        if title not in self.tables.keys():
            table=Table(title=title,box=box.SIMPLE_HEAD,title_style="bold black on white")
            table.add_column('选项',justify="left")
            table.add_column('功能',justify="left")
            for idx,text in opts.items():
                table.add_row(f'{idx: >2}. ',f'{text}')
            self.tables[title]=table
        table=self.tables[title]
        self.console.print(table)
    
    def track(self,seq:Sequence,tip:str='')->Iterable:
        if not config.IF_SHELL:return
        from rich import progress
        return progress.track(seq,description=tip)

printer=Printer()

"""不需要多次实例化的类可以使用类方法"""
import time
class TimeCounter:
    """统计每个函数执行的时间"""
    counts={}
    
    def __init__(self) -> None:
        pass
    
    @classmethod
    def funcTime(cls,func):
        t0=time.time()
        ans=func()
        t1=time.time()
        
        return ans

def vectStr(vector:None|np.ndarray,f=4):
    if vector is None:
        return 'None'
    else:
        resStr=','.join([f'{v:>8.4f}' for v in vector])
        return f'({resStr})'


"""
单点能与自由能能量矫正
仅能使用log文件！
"""
from pywfn.reader import LogReader

from pathlib import Path
import numpy as np

class Tool:
    """能量矫正工具"""
    def __init__(self,sfold:str,ffold:str) -> None:
        """
        sfold：single point energy folder 单点能文件所在目录
        ffold：free energy folder 自由能文件所在目录
        """
        self.sfold=sfold
        self.ffold=ffold
        self.sengs=[]
        self.fengs=[]
        self.files=[]

    def extract(self):
        """循环从文件中提取能量"""
        for path in Path(self.sfold).iterdir():
            if path.suffix!='.log':continue
            seng=LogReader(f'{path}').read_energy() # 计算单点能
            self.sengs.append(seng)
            self.files.append(path.name)
        
        for path in Path(self.ffold).iterdir():
            if path.suffix!='.log':continue
            name,engs=LogReader(f'{path}').read_energys()
            feng=engs[3]
            self.fengs.append(feng)

    def save(self,path:str):
        """保存到指定文件"""
        self.extract()
        resStr='file,seng,feng,sum\n'
        for file,seng,feng in zip(self.files,self.sengs,self.fengs):
            resStr+=f'{file},{seng},{feng},{seng+feng}\n'
        Path(path).write_text(resStr)


"""
本脚本用来读取.xyz文件
xyz文件只包含原子坐标,所以很好读
一个xyz文件可以包含多个分子的坐标,所以要生成多个分子对象
"""
from typing import List
import re

from pywfn import base
from pywfn.reader.lutils import Reader

class XyzReader(Reader):
    def __init__(self,path) -> None:
        self.path=path
        with open(path,'r',encoding='utf-8') as f:
            self.content=f.read()
        self.logLines=self.content.splitlines(keepends=False)
        self.mols:List["base.Mol"]=[]
        self.read_coord()
    
    def read_coord(self):
        """
        有三种情况,原子数量,标题,原子坐标
        """
        
        s1='^ +\d+$'
        # 不满足第一种和第二种情况就是第三种情况
        s3='^ +([A-Za-z])+ +(-?\d+.\d+) +(-?\d+.\d+) +(-?\d+.\d+)$'

        for line in self.logLines:
            if re.search(s1,line) is not None:
                self.mols.append(base.Mol())
            elif re.search(s3,line) is not None:
                symbol,x,y,z=re.search(s3,line).groups()
                self.mols[-1].add_atom(symbol,[float(x),float(y),float(z)])
            else:
                continue



"""
实现单位转换
既然作为波函数分析的程序，针对的都是原子和分子，那么使用原子单位就比较合适。
使用原子单位 https://zhuanlan.zhihu.com/p/362046144
"""

# 定义支持的单位和类型，同一个类型的单位之间可以互相转换

class Unit:
    def __init__(self,utype:str) -> None:
        self.utype=utype # 单位类型
        self.maps:dict[str,float]={} # 单位与转换系数之间的映射
        self.value:str
    
    @property
    def stand(self)->str:
        return self.units[0]

    @property
    def units(self)->list[str]:
        return list(self.maps.keys())
    
    @property
    def coeff(self)->list[float]:
        return list(self.maps.values())
    
    def set(self,unit:str,value:float):
        assert self.include(unit),f'不符合的单位{unit}'
        index=self.units.index(unit)
        self.value=value/self.coeff[index]
        return self
    
    def get(self,unit:str): # 将指定单位的数值转为标准单位
        assert self.include(unit),f'不符合的单位{unit}'
        index=self.units.index(unit)
        return self.value*self.coeff[index]
    
    def include(self,unit:str):
        if unit in self.units:
            return True
        return False

class ChargeUnit(Unit):
    def __init__(self) -> None:
        super().__init__('charge')
        self.maps={
            'e':1.,
            'C':1.602e-19
        }

class LengthUnit(Unit):
    def __init__(self) -> None:
        super().__init__('length')
        self.maps={
            'bohr':1.,
            'angstrom':0.52918,
            'A':0.52918,
            'nm':0.1
        }
        
class EnergyUnit(Unit):
    def __init__(self) -> None:
        super().__init__('energy')
        self.maps={
            'hartree':1.,
            'eV':27.211,
            'KJ/mol':2625,
            'J':1.36e-18
        }

units:list[Unit]=[ChargeUnit(),LengthUnit(),EnergyUnit()]

def trans(unit0:str,unit1:str,value:float):
    """将单位0转换为单位1"""
    for unit in units:
        if unit.include(unit0):
            assert unit.include(unit1),f'不符合的单位{unit1}'
            return unit.set(unit0,value).get(unit1)
    raise ValueError(f'不符合的单位{unit0}')

def coeff(unit0:str,unit1:str):
    """获取从单位0到单位1的转换系数"""
    for unit in units:
        if unit.include(unit0):
            assert unit.include(unit1),f'不符合的单位{unit1}'
            return unit.maps[unit1]/unit.maps[unit0]
    raise ValueError(f'不符合的单位{unit0}')


"""
计算原子自由价
原子自由价=最大键级-该原子键级之和
这里的键级就先用DH
"""
from pywfn.bondprop import piDM
from pywfn.base import Mol,Atom
from pywfn.utils import printer
from pywfn.atomprop import AtomCaler
import numpy as np

STAND=1.6494

class Calculator(AtomCaler):
    def __init__(self,mol:Mol) -> None:
        self.mol=mol
        self.caler=piDM.Calculator(self.mol)
        self.vect:np.ndarray=None
        self.atom:int=None
        self.zero:bool=True
        self.keep:bool=True
        self.ins:bool=True
    
    def calculate(self):
        """计算一个原子的自由价，从1开始"""
        assert self.vect is not None,'未指定方向'
        atm=self.atom
        printer.info(f'计算原子 {atm} 自由价')
        centAtom=self.mol.atom(atm)
        valence1=STAND
        valence2=STAND
        self.caler.vect=self.vect.copy()
        self.caler.zero=self.zero
        self.caler.keep=self.keep
        self.caler.ins=self.ins

        for arouAtm in centAtom.neighbors:
            self.caler.bond=[atm,arouAtm]
            orders=self.caler.calculate()
            order1,order2=orders
            bond=f'{centAtom.idx}-{arouAtm}'
            printer.log(f'{bond} {order1:.6f} {order2:.6f}')
            valence1-=order1
            valence2-=order2
        if valence2==STAND:valence2=0.0
        return valence1,valence2
    
    
    def print(self,resStr:str):
        printer.info('原子自由价及其相关键级: ')
        printer.res(resStr)
    
    def resStr(self,idxs:list[int]):
        """结果的字符串形式"""
        valences=self.calculate(idxs)
        resStr=''
        for atom,(value1,value2) in zip(idxs,valences):
            resStr+=f'{atom:<8}{value1:<10.6f}{value2:<10.6f}\n'
        return resStr
